        TITLE   vmp3DX86.asm - 80(x)86 3D Math
        PAGE    53,160

;  vmp3DX86.asm - 3D Math
;
;  This document contains file that you may use at your own risk in your own programs.
;
;  This module is designed to accomodate the book.
;
;               This module is designed for fast 3D math
;
;  Created - 3 Oct 2001 - James Leiterman
;
;  Last Modification
;
;	23 Apr 2008 - James Leiterman (Compatibility with vc 2005)
;       Tabs = 8



        .XLIST

;;;        .387
;;;     .586P   .586    .686   .686P
;;;      .K3D    .XMM    .MMX

        .686
        .XMM

        .model flat, C

;ifdef _DEBUG
USE_ASSERT      =       1
;endif

        include KariType.inc
        include vmp3DX86M.inc
 
        .LIST



        .data
        align   16


signMask        qword   07fffffff7fffffffh,07fffffff7fffffffh
;;;zero         oword   0       ;,0     ;REAL4  0,0,0,0


scratch dword   0,0,0,0, 0,0,0,0,  0,0,0,0,  0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0



;------------------------------------------------------------------------------------------------


vmp3DX86        SEGMENT USE32 PUBLIC 'CODE'

;-------------------------------------------------------------------------------------------------


xmmDump proc    near
        movaps  scratch+00h,xmm0
        movaps  scratch+10h,xmm1
        movaps  scratch+20h,xmm2
        movaps  scratch+30h,xmm3
        movaps  scratch+40h,xmm4
        movaps  scratch+50h,xmm5
        movaps  scratch+60h,xmm6
        movaps  scratch+70h,xmm7
        ret
xmmDump endp





;_________________________________________________________________________________________________
;=================================================================================================
;
;bool vmp_IsQVEqual( const vmp3DQVector * const pvA, const vmp3DQVector * const pvB, float fPrec )
;{
;       if (   !vmp_IsFEqual(pvA->x, pvB->x, fPrec) 
;               || !vmp_IsFEqual(pvA->y, pvB->y, fPrec) 
;               || !vmp_IsFEqual(pvA->z, pvB->z, fPrec) 
;               || !vmp_IsFEqual(pvA->w, pvB->w, fPrec))
;         {
;               return false;
;         }
;
;       return true;
;}                                                                                              
;
;_________________________________________________________________________________________________

        public  vmp_IsQVEqualAsm3DNow
vmp_IsQVEqualAsm3DNow   proc    near
        .K3D

ASSERT_PTR4( DWORD PTR [esp+8] )                ; bMx
ASSERT_PTR4( DWORD PTR [esp+4] )                ; aMx

        movd    mm7,(DWORD PTR [esp+12])            ; fPrecision
        mov     ecx,[esp+8]             ; bMx
        mov     eax,[esp+4]             ; aMx

        punpckldq mm7,mm7                       ; {fPrec fPrec}

                ;       1st vector

        movq    mm0,[eax+(0*4)]         ; {a1 a0}
        movq    mm2,[ecx+(0*4)]         ; {b1 b0}
        movq    mm1,[eax+(2*4)]         ; {a3 a2}
        movq    mm3,[ecx+(2*4)]         ; {b3 b2}

        pand    mm7,(QWORD PTR signMask)            ; Strip sign bit

        pfsub   mm0,mm2
        pfsub   mm1,mm3

        pand    mm0,(QWORD PTR signMask)            ; Strip sign bit
        pand    mm1,(QWORD PTR signMask)            ; Strip sign bit
        
        pfcmpgt mm0,mm7                 ; (n > precision) ? -1 : 0
        pfcmpgt mm1,mm7                 ; (n > precision) ? -1 : 0

                ;       Merge results, if -1 then one field was greater!

        por     mm0,mm1                 ; [63...32 31...0]

        pswapd  mm1,mm0                 ; [31...0  63...32]
        por     mm0,mm1

        movd    eax,mm0                 ; -1 if NOT equal
        not     eax
        shr     eax,31                  ; 1 is equal, 0 is not!
        ret
vmp_IsQVEqualAsm3DNow   endp



;_________________________________________________________________________________________________

        public  vmp_IsQVEqualAsmSSE
vmp_IsQVEqualAsmSSE     proc    near
        .XMM

ASSERT_PTR16( DWORD PTR [esp+8] )               ; pvB
ASSERT_PTR16( DWORD PTR [esp+4] )               ; pvA

        movss   xmm7,(DWORD PTR [esp+12])           ; fPrecision
        mov     ecx,[esp+8]             ; bMx
        mov     eax,[esp+4]             ; aMx

        movaps  xmm6,(OWORD PTR signMask)
        pshufd  xmm7,xmm7,00000000b     ; {fPrec fPrec fPrec fPrec}

                ;       Upper 2 vectors

        movaps  xmm0,[eax+(0*4)]        ; {a3 a2 a1 a0}
        movaps  xmm1,[ecx+(0*4)]        ; {b3 b2 b1 b0}

        andps   xmm7,xmm6               ; Strip sign bit

        subps   xmm0,xmm1               ; [a3-b3 a2-b2 a1-b1 a0-b0]
        andps   xmm0,xmm6               ; Strip sign bit
        cmpps   xmm0,xmm7,2             ; (n <= precision) ? 0 : -1

        ; SSE doesn't support bit twiddling, so get 'xmm0' into MMX into EAX as a 1:0
        ; THE HARD WAY!!!

        movups  [esp-20],xmm0           ; Use stack as a scratch pad
        movq    mm0,[esp-20]
        movq    mm1,[esp-12]

        pand    mm0,mm1                 ;{63...32 31...0}
        punpckhdq       mm1,mm1         ;{63...32 63...32}
        pand    mm0,mm1

        movd    eax,mm0
        shr     eax,31                  ; 1 is equal, 0 is not!
        ret
vmp_IsQVEqualAsmSSE     endp


;_________________________________________________________________________________________________
;=================================================================================================
;
;bool vmp_IsMatrixEqual( vmp3DMatrix aMx, vmp3DMatrix bMx, float fPrecision)
;       uint u;
;       float *pfA, *pfB;
;       
;       pfA = (float *) aMx;
;       pfB = (float *) bMx;
;       
;       u = 16;
;       while (u--)
;         {
;               if (!vmp_IsFEqual( *pfA++, *pfB++, fPrecision ))
;                 {
;                       return false;
;                 }
;         }
;
;       return true;
;_________________________________________________________________________________________________

        public  vmp_IsMatrixEqualAsm3DNow
vmp_IsMatrixEqualAsm3DNow       proc    near
        .K3D

ASSERT_PTR4( DWORD PTR [esp+8] )                ; bMx
ASSERT_PTR4( DWORD PTR [esp+4] )                ; aMx

        movd    mm7,(DWORD PTR [esp+12])            ; fPrecision
        mov     ecx,[esp+8]             ; bMx
        mov     eax,[esp+4]             ; aMx

        punpckldq mm7,mm7                       ; {fPrec fPrec}

                ;       1st vector

        movq    mm0,[eax+(0*4)]         ; {a1 a0}
        movq    mm2,[ecx+(0*4)]         ; {b1 b0}
        movq    mm1,[eax+(2*4)]         ; {a3 a2}
        movq    mm3,[ecx+(2*4)]         ; {b3 b2}

        pand    mm7,(QWORD PTR signMask)            ; Strip sign bit

        pfsub   mm0,mm2
        pfsub   mm1,mm3

        pand    mm0,(QWORD PTR signMask)            ; Strip sign bit
        pand    mm1,(QWORD PTR signMask)            ; Strip sign bit
        
        pfcmpgt mm0,mm7                 ; (n > precision) ? -1 : 0
        pfcmpgt mm1,mm7                 ; (n > precision) ? -1 : 0

i = 4
REPEAT 3
                ;       2nd vector

        movq    mm2,[eax+((i+0)*4)]             ; {b5 b4}
        movq    mm4,[ecx+((i+0)*4)]             ; {a5 a4}
        movq    mm3,[eax+((i+2)*4)]             ; {b7 b6}
        movq    mm5,[ecx+((i+2)*4)]             ; {a7 a6}

        pfsub   mm2,mm4
        pfsub   mm3,mm5

        pand    mm2,signMask            ; Strip sign bit
        pand    mm3,signMask            ; Strip sign bit
        
        pfcmpgt mm2,mm7                 ; (n > precision) ? -1 : 0
        pfcmpgt mm3,mm7                 ; (n > precision) ? -1 : 0

        por     mm0,mm2                 ; Blend bits, if one element > then mark as not equal!
        por     mm1,mm3
i = i + 4
	ENDM
                ;       Merge results, if -1 then one field was greater!

        por     mm0,mm1                 ; [63...32 31...0]

        pswapd  mm1,mm0                 ; [31...0  63...32]
        por     mm0,mm1

        movd    eax,mm0                 ; -1 if NOT equal
        not     eax
        shr     eax,31                  ; 1 is equal, 0 is not!
        ret
vmp_IsMatrixEqualAsm3DNow       endp



;_________________________________________________________________________________________________

        public  vmp_IsMatrixEqualAsmSSE
vmp_IsMatrixEqualAsmSSE proc    near
        .XMM

ASSERT_PTR16( DWORD PTR [esp+8] )               ; bMx
ASSERT_PTR16( DWORD PTR [esp+4] )               ; aMx

        movss   xmm7,(DWORD PTR [esp+12])           ; fPrecision
        mov     ecx,[esp+8]             ; bMx
        mov     eax,[esp+4]             ; aMx

        movaps  xmm6,(OWORD PTR signMask)
        pshufd  xmm7,xmm7,00000000b     ; {fPrec fPrec fPrec fPrec}

                ;       Upper 2 vectors

        movaps  xmm0,[eax+(0*4)]        ; {a3 a2 a1 a0}
        movaps  xmm2,[ecx+(0*4)]        ; {b3 b2 b1 b0}
        movaps  xmm1,[eax+(4*4)]        ; {a7 a6 a5 a4}
        movaps  xmm3,[ecx+(4*4)]        ; {b7 b6 b5 b4}

        andps   xmm7,xmm6               ; Strip sign bit

        subps   xmm0,xmm2               ; [a3-b3 a2-b2 a1-b1 a0-b0]
        subps   xmm1,xmm3               ; [a7-b7 a6-b6 a5-b5 a4-b4]

        andps   xmm0,xmm6               ; Strip sign bit
        andps   xmm1,xmm6
        
        cmpps   xmm0,xmm7,2             ; (n <= precision) ? 0 : -1
        cmpps   xmm1,xmm7,2             ; (n <= precision) ? 0 : -1

                ;       Lower 2 vectors

        movaps  xmm2,[eax+(8*4)]        ; {aB aA a9 a8}
        movaps  xmm4,[ecx+(8*4)]        ; {bB bA b9 b8}
        movaps  xmm3,[eax+(12*4)]       ; {aF aE aD aC}
        movaps  xmm5,[ecx+(12*4)]       ; {bF bE bD bC}

        subps   xmm2,xmm4               ; [aB-bB aA-bA a9-b9 a8-b8]
        subps   xmm3,xmm5               ; [aF-bF aE-bE aD-bD aC-bC]

        andps   xmm2,xmm6               ; Strip sign bit
        andps   xmm3,xmm6
        
        cmpps   xmm2,xmm7,2             ; (n <= precision) ? 0 : -1
        cmpps   xmm3,xmm7,2             ; (n <= precision) ? 0 : -1

        ;       Bit Mixing - All Bits are 1 if equal!  xmm0...xmm3

        andps   xmm0,xmm1
        andps   xmm2,xmm3
        andps   xmm0,xmm2

        ; SSE doesn't support bit twiddling, so get 'xmm0' into MMX into EAX as a 1:0
        ; THE HARD WAY!!!

        movups  [esp-20],xmm0           ; Use stack as a scratch pad
        movq    mm0,[esp-20]
        movq    mm1,[esp-12]

        pand    mm0,mm1                 ;{63...32 31...0}
        punpckhdq       mm1,mm1         ;{63...32 63...32}
        pand    mm0,mm1

        movd    eax,mm0
        shr     eax,31                  ; 1 is equal, 0 is not!
        ret
vmp_IsMatrixEqualAsmSSE endp



vmp3DX86        ends
        end

